@@ -1,6 +1,6 @@
 package Setup::File::Symlink;
 BEGIN {
-  $Setup::File::Symlink::VERSION = '0.05';
+  $Setup::File::Symlink::VERSION = '0.06';
 }
 # ABSTRACT: Ensure symlink existence and target
 
@@ -9,6 +9,11 @@ use strict;
 use warnings;
 use Log::Any '$log';
 
+use Data::Dump::Partial qw(dumpp);
+use File::Copy::Recursive qw(rmove);
+use File::Path qw(remove_tree);
+use UUID::Random;
+
 require Exporter;
 our @ISA       = qw(Exporter);
 our @EXPORT_OK = qw(setup_symlink);
@@ -17,6 +22,21 @@ our %SPEC;
 
 $SPEC{setup_symlink} = {
     summary  => "Create symlink or fix symlink target",
+    description => <<'_',
+
+On do, will create symlink which points to specified target. If symlink already
+exists but points to another target, it will be replaced with the correct
+symlink if replace_symlink option is true. If a file already exists, it will be
+removed (or, backed up to temporary directory) before the symlink is created, if
+replace_file option is true.
+
+If given, -undo_hint should contain {tmp_dir=>...} to specify temporary
+directory to save replaced file/dir. Temporary directory defaults to ~/.setup,
+it will be created if not exists.
+
+On undo, will restore the original symlink/ if it was replaced during do.
+
+_
     args     => {
         symlink => ['str*' => {
             summary => 'Path to symlink',
@@ -57,8 +77,6 @@ _
 
 If set to false, then setup will fail (412) if this condition is encountered.
 
-NOTE: Not yet implemented, so will always fail under this condition.
-
 _
             default => 0,
         }],
@@ -68,8 +86,6 @@ _
 
 If set to false, then setup will fail (412) if this condition is encountered.
 
-NOTE: Not yet implemented, so will always fail under this condition.
-
 _
             default => 0,
         }],
@@ -88,71 +104,129 @@ sub setup_symlink {
     my $target  = $args{target};
     defined($target) or return [400, "Please specify target"];
     my $create       = $args{create} // 1;
-    my $replace_file = $args{replace_dir} // 0;
-    my $replace_dir  = $args{replace_file} // 0;
+    my $replace_file = $args{replace_file} // 0;
+    my $replace_dir  = $args{replace_dir} // 0;
     my $replace_sym  = $args{replace_symlink} // 1;
 
-    # check current state
+    # check current state and collect steps
     my $is_symlink = (-l $symlink); # -l performs lstat()
     my $exists     = (-e _);        # now we can use -e
     my $is_dir     = (-d _);
     my $cur_target = $is_symlink ? readlink($symlink) : "";
-    my $state_ok   = $is_symlink && $cur_target eq $target;
+    my $steps = [];
+    if ($exists && !$is_symlink) {
+        $log->tracef("nok: exist but not a symlink");
+        if ($is_dir) {
+            if (!$replace_dir) {
+                return [412, "must replace dir but instructed not to"];
+            }
+            push @$steps, ["rm"], ["ln"];
+        } else {
+            if (!$replace_file) {
+                return [412, "must replace file but instructed not to"];
+            }
+            push @$steps, ["rm"], ["ln"];
+        }
+    } elsif ($is_symlink && $cur_target ne $target) {
+        $log->tracef("nok: symlink doesn't point to correct target");
+        if (!$replace_sym) {
+            return [412, "must replace symlink but instructed not to"];
+        }
+        push @$steps, ["rmsym"], ["ln"];
+    } elsif (!$exists) {
+        $log->tracef("nok: doesn't exist");
+        if (!$create) {
+            return [412, "must create symlink but instructed not to"];
+        }
+        push @$steps, ["ln"];
+    }
 
+    # perform the steps
     if ($undo_action eq 'undo') {
         return [412, "Can't undo: currently $symlink is not a symlink ".
-                    "pointing to $target"] unless $state_ok;
-        return [304, "dry run"] if $dry_run;
-        unlink $symlink
-            or return [500, "Can't undo: unlink $symlink: $!"];
-        my $undo_data = $args{-undo_data};
-        if ($undo_data->[0] eq 'dir') {
-            # XXX mv $undo_data->[1], $symlink;
-        } elsif ($undo_data->[0] eq 'file') {
-            # XXX mv $undo_data->[1], $symlink;
-        } elsif ($undo_data->[0] eq 'symlink') {
-            $log->tracef("undo setup_symlink: restoring old symlink %s -> %s",
-                         $symlink, $undo_data->[1]);
-            return [304, "dry run"] if $dry_run;
-            symlink $undo_data->[1], $symlink
-                or return [500, "Can't undo: symlink $symlink -> $target: $!"];
-        } elsif ($undo_data->[0] eq 'none') {
-            $log->tracef("undo setup_symlink: deleting symlink %s", $symlink);
+                    "pointing to $target"] unless !@$steps;
+        $steps = $args{-undo_data} or return [400, "Please supply -undo_data"];
+    } elsif ($undo_action eq 'redo') {
+        $steps = $args{-redo_data} or return [400, "Please supply -redo_data"];
+    }
+
+    return [400, "Invalid steps, must be an array"]
+        unless $steps && ref($steps) eq 'ARRAY';
+
+    return [304, "Nothing to do"] unless @$steps;
+    return [200, "Dry run"] if $dry_run;
+
+    # create tmp dir for undo
+    my $save_undo    = $undo_action ? 1:0;
+    my $undo_hint = $args{-undo_hint} // {};
+    return [400, "Invalid -undo_hint, please supply a hashref"]
+        unless ref($undo_hint) eq 'HASH';
+    my $tmp_dir = $undo_hint->{tmp_dir} // "$ENV{HOME}/.setup";
+    if ($save_undo && !(-d $tmp_dir)) {
+        mkdir $tmp_dir or return [500, "Can't make temp dir `$tmp_dir`: $!"];
+    }
+    my $save_path = "$tmp_dir/".UUID::Random::generate;
+
+    my $is_rollback;
+    my $undo_steps = [];
+  STEPS:
+    for my $i (0..@$steps-1) {
+        my $step = $steps->[$i];
+        $log->tracef("step %d of 0..%d: %s", $i, @$steps-1, $step);
+        my $err;
+        return [400, "Invalid step (not array)"] unless ref($step) eq 'ARRAY';
+        if ($step->[0] eq 'rmsym') {
+            if (unlink $symlink) {
+                unshift @$undo_steps, ["ln", $cur_target];
+            } else {
+                $err = "Can't remove $symlink: $!";
+            }
+        } elsif ($step->[0] eq 'rm') {
+            # do not bother to save file/dir if not asked
+            if ($save_undo) {
+                if (rmove $symlink, $save_path) {
+                    unshift @$undo_steps, ["restore", $save_path];
+                } else {
+                    $err = "Can't move file/dir $symlink -> $save_path: $!";
+                }
+            } else {
+                remove_tree($symlink, {error=>\my $e});
+                if (@$e) {
+                    $err = "Can't remove file/dir $symlink: ".dumpp($e);
+                }
+            }
+        } elsif ($step->[0] eq 'restore') {
+            if (rmove $step->[1], $symlink) {
+                unshift @$undo_steps, ["rm"];
+            } else {
+                $err = "Can't restore $step->[1] -> $symlink: $!";
+            }
+        } elsif ($step->[0] eq 'ln') {
+            my $t = $step->[1] // $target;
+            if (symlink $t, $symlink) {
+                unshift @$undo_steps, ["rmsym"];
+            } else {
+                $err = "Can't symlink $symlink -> $target: $!";
+            }
         } else {
-            return [412, "Invalid undo data"];
+            die "BUG: Unknown step command: $step->[0]";
+        }
+        if ($err) {
+            if ($is_rollback) {
+                die "Failed rollback step $i of 0..".(@$steps-1).": $err";
+            } else {
+                $log->tracef("Step failed: $err, performing rollback ...");
+                $is_rollback++;
+                $steps = $undo_steps;
+                redo STEPS;
+            }
         }
-        return [200, "OK", undef, {}];
     }
 
-    my $undo_hint = $args{-undo_hint};
-
-    if ($state_ok) {
-        return [304, "Already ok"];
-    } elsif (!$exists) {
-        return [412, "Should create but told not to"] unless $create;
-        $log->tracef("setup_symlink: creating symlink %s", $symlink);
-        return [304, "dry run"] if $dry_run;
-        symlink $target, $symlink or return [500, "Can't symlink: $!"];
-        return [200, "Created", undef, {undo_data=>['none']}];
-    } elsif ($is_symlink) {
-        return [412, "Should replace symlink but told not to, ".
-                    "please delete $symlink manually first"]
-            unless $replace_sym;
-        $log->tracef("setup_symlink: replacing symlink %s", $symlink);
-        return [304, "dry run"] if $dry_run;
-        unlink $symlink or return [500, "Can't unlink $symlink: $!"];
-        symlink $target, $symlink or return [500, "Can't symlink: $!"];
-        return [200, "Replaced symlink", undef,
-                {undo_data=>[symlink=>$cur_target]}];
-    } elsif ($is_dir) {
-        return [412, "Can't setup symlink $symlink because it is currently ".
-            "a dir, please delete it manually first"];
-        # XXX
-    } else {
-        return [412, "Can't setup symlink $symlink because it is currently ".
-            "a file, please delete it manually first"];
-        # XXX
-    }
+    my $meta = {};
+    if ($undo_action =~ /^(re)?do$/) { $meta->{undo_data} = $undo_steps }
+    elsif ($undo_action eq 'undo')   { $meta->{redo_data} = $undo_steps }
+    return [200, "OK", undef, $meta];
 }
 
 1;
@@ -166,7 +240,7 @@ Setup::File::Symlink - Ensure symlink existence and target
 
 =head1 VERSION
 
-version 0.05
+version 0.06
 
 =head1 SYNOPSIS
 
@@ -174,18 +248,24 @@ version 0.05
 
  # simple usage (doesn't save undo data)
  my $res = setup_symlink symlink => "/baz", target => "/qux";
- die unless $res->[0] == 200;
+ die unless $res->[0] == 200 || $res->[0] == 304;
 
- # perform setup and save undo data (undo data should be serializable)
+ # perform setup and save undo data
  my $res = setup_symlink symlink => "/foo", target => "/bar",
                          -undo_action => 'do';
- die unless $res->[0] == 200;
+ die unless $res->[0] == 200 || $res->[0] == 304;
  my $undo_data = $res->[3]{undo_data};
 
  # perform undo
  my $res = setup_symlink symlink => "/symlink", target=>"/target",
                          -undo_action => "undo", -undo_data=>$undo_data;
  die unless $res->[0] == 200;
+ my $redo_data = $res->[3]{redo_data};
+
+ # perform redo
+ my $res = setup_symlink symlink => "/symlink", target=>"/target",
+                         -undo_action => "redo", -redo_data=>$redo_data;
+ die unless $res->[0] == 200;
 
 =head1 DESCRIPTION
 
@@ -213,8 +293,12 @@ correct content/ownership/permission.
 
 =item * do nothing if desired state has been reached
 
+Function should return 304 (nothing to do) status.
+
 =item * support dry-run (simulation) mode
 
+Function should return 200 on success, but change nothing.
+
 =item * support undo to restore state to previous/original one
 
 =back
@@ -228,6 +312,18 @@ None are exported by default, but they are exportable.
 
 Create symlink or fix symlink target.
 
+On do, will create symlink which points to specified target. If symlink already
+exists but points to another target, it will be replaced with the correct
+symlink if replace_symlink option is true. If a file already exists, it will be
+removed (or, backed up to temporary directory) before the symlink is created, if
+replace_file option is true.
+
+If given, -undo_hint should contain {tmp_dir=>...} to specify temporary
+directory to save replaced file/dir. Temporary directory defaults to ~/.setup,
+it will be created if not exists.
+
+On undo, will restore the original symlink/ if it was replaced during do.
+
 Returns a 3-element arrayref. STATUS_CODE is 200 on success, or an error code
 between 3xx-5xx (just like in HTTP). ERR_MSG is a string containing error
 message, RESULT is the actual result.
@@ -264,16 +360,12 @@ Replace if there is existing dir.
 
 If set to false, then setup will fail (412) if this condition is encountered.
 
-NOTE: Not yet implemented, so will always fail under this condition.
-
 =item * B<replace_file>* => I<bool> (default C<0>)
 
 Replace if there is existing non-symlink file.
 
 If set to false, then setup will fail (412) if this condition is encountered.
 
-NOTE: Not yet implemented, so will always fail under this condition.
-
 =item * B<replace_symlink>* => I<bool> (default C<1>)
 
 Replace previous symlink if it already exists but doesn't point to the wanted target.
